/* 
 * Copyright (C) 2005 Dusan Gallo
 * Email: dusky@hq.alert.sk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. See the file COPYING. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * divIDE extension (http://baze.au.com/divide)
 *
 * I would like to thank to Zilog, baze, nairam for their useful 
 * comments/informations. dusky
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include "z80_type.h"
#include "z80.h"
#include "divide.h"
#include "interf.h"

#define DISK divide_drive[drive]

// DEMFIR operating system for divide (http://demfir.sourceforge.net)
extern unsigned char demfir_imag[];

// mapped original ROM or divide memory?
int divide_memstate=0;

// forced memory mapping
int divide_conmem=0;

// RAM page 3 mapped RO in 0K-8K range
int divide_mapram=0;

// flash write jumper
int divide_jp2=1;

// which memry page memory in 8k-16k
int divide_bank=0;
int divide_bank_changed=0;

// if should be done automapping
int divide_automap=0;

// IDE registers (Read/Write)
int divide_port_AF;	// chs/lba
int divide_port_B3;	// chs/lba
int divide_port_B7;	// chs/lba
int divide_port_BB;	// chs/lba
int divide_port_A3;	// data
int divide_port_AB;	// sector cnt

// IDE registers (Write)
int divide_out_port_A7;	// features
int divide_out_port_BF;	// command

// IDE registers (Read)
int divide_in_port_A7;	// error
int divide_in_port_BF;	// status

#define DIVIDE_MAX_SEC_NR 96 // 96*512 = 48K .. in 48K mode spectrum this should be enough

// active drive (0='A:', 1='B:')
int divide_active_drive=0;

struct divide_drive_struct {
  char *path;		// image path
  int c,h,s;		// C/H/S geometry of disk (not used yet)
  int len;		// size of image (readed from identify file)
  int ptr;		// head ptr (not used)

  byte identify[512];	// identify virtual sector

// See README.divIDE for more information
  int mbr_fd;		// Master Boot Record info
  int mbr_len;
  int mbr_ptr;

  int partition_fd;	// partition info
  int partition_len;
  int partition_ptr;
//

  byte buffer[512*DIVIDE_MAX_SEC_NR];
  int buffer_ptr;
  int buffer_len;

  int cmd;		// pending command (read sector/write sector)

};

struct divide_drive_struct divide_drive[2];

#define DIVIDE_READ_BUFF	1
#define DIVIDE_WRITE_BUFF	2

// due to not_so_good memory structure used in spectemu i had to switch
// memory by copying it, maybe i would rewrite it later
byte orig_rom[16384];		
byte divide_eeprom[8192];
byte divide_ram[4][8192];

// C/H/S geometry -> LBA
int chs2lba(int drive, int c, int h, int s) {
  return (s+DISK.s*(h+c*DISK.h)-1);
}

// initialize variables
void divide_init() {
  int i,j;
  
  // make copy of original spectrum memory (to recover it after unmaping divide memory)
  memcpy(orig_rom, PRNM(proc).mem, 16384);
  
  // load EEPROM memory and clear RAM memory
  for(i = 0; i < 0x2000; i++)
      divide_eeprom[i] = demfir_imag[i];
  for(i = 0; i < 0x2000; i++)
    for (j=0;j<4;j++)
        divide_ram[j][i] = 0;

  // set default values
  divide_port_A3=0;
  divide_port_AB=0;
  divide_port_AF=0;
  divide_port_B3=0;
  divide_port_B7=0;
  divide_port_BB=0;

  divide_out_port_A7=0;
  divide_out_port_BF=0;
  
  divide_in_port_A7=0;
  divide_in_port_BF=0;
  divide_drive[0].path=NULL;
  divide_drive[0].mbr_fd=-1;
  divide_drive[0].partition_fd=-1;
  divide_drive[1].path=NULL;
  divide_drive[1].mbr_fd=-1;
  divide_drive[1].partition_fd=-1;
}

// divide images are directories in fact
// it must contain files:
//   .identify (virtual sector containing drive firmware information)
//   .mbr (master boot record [and rest of track)
//   part1 (partition image)
//   part2..partN, .ebr (other partitions, not implemented in this stage)
// this model was choosed to easy mount divide images under linux

void divide_attach_drive(int drive, char *path) {
  char *fname;
  int fd, r, len;
  byte identify[512];
  
  if (!path) return;
  
  // allocate buffer for filenames
  len=strlen(path) + 15;
  fname=malloc(sizeof(char) * (len + 1));
  if (!fname) return;
  fname[len]='\0';
  
  // read contents of .identify "sector" if it doesn't exists ignore image
  snprintf(fname, len, "%s/.identify", path);
  if ((fd=open(fname, O_RDONLY))!=-1) {
    r=read(fd, identify, 512);
    close(fd);
    if (r!=512) return;
    
    // if another drive was mounted
    if ((DISK.path)) divide_detach_drive(drive);
    if (!(DISK.path=strdup(path))) return;
    
    memcpy(DISK.identify, identify, 512);
    
    // read Master Boot Record information
    snprintf(fname, len, "%s/.mbr", path);
    if ((DISK.mbr_fd=open(fname, O_RDWR))!=-1) {
      // length is in sectors (512bytes blocks), extra chunk is ignored (if size is not divisible by 512)
      DISK.mbr_len=(lseek(DISK.mbr_fd, 0, SEEK_END))>>9;
      DISK.mbr_ptr=0;
      // this is for realistic emulation of moving disk head (not used now)
      lseek(DISK.mbr_fd, DISK.mbr_ptr, SEEK_SET);
    }
    
    // read first partition
    snprintf(fname, len, "%s/part1", path);
    if ((DISK.partition_fd=open(fname, O_RDWR))!=-1) {
      // length is in sectors (512bytes blocks), extra chunk is ignored (if size is not divisible by 512)
      DISK.partition_len=(lseek(DISK.partition_fd, 0, SEEK_END))>>9;
      DISK.partition_ptr=0;
      // this is for realistic emulation of moving disk head (not used now)
      lseek(DISK.partition_fd, DISK.partition_ptr, SEEK_SET);
    }

    DISK.buffer_ptr=0;
    DISK.buffer_len=0;
    DISK.cmd=0;
    
    // C/H/S configuration from identify "sector"
    DISK.c=identify[108]+256*identify[109];
    DISK.h=identify[110]+256*identify[111];
    DISK.s=identify[112]+256*identify[113];
    
    if (!drive)
      put_msg("Image successfully attached as drive 'A:'.");
    else
      put_msg("Image successfully attached as drive 'B:'.");
  } else put_msg("Image not found.");
  free(fname);
}

void divide_detach_drive(int drive) {

  if (DISK.path) {
    free(DISK.path);
    DISK.path=NULL;
  }
  if (DISK.mbr_fd!=-1) close(DISK.mbr_fd);
  if (DISK.partition_fd!=-1) close(DISK.partition_fd);
  if (!drive)
    put_msg("Image successfully detached from drive 'A:'.");
  else
    put_msg("Image successfully detached from drive 'B:'.");
}

void divide_select_drive(int drive) {
  char *name;

  if (!drive)
    put_msg("Enter path to image (drive 'A:'):");
  else
    put_msg("Enter path to image (drive 'B:'):");

  name = spif_get_filename();
  if(name == NULL) return;

  divide_detach_drive(drive);
  divide_attach_drive(drive, name);
}

// Well.. ready to go to the jungle? ;]

// IN wrapper
int divide_port_in(int portl) {
  int data, drive;

  drive = divide_active_drive;
  switch (portl) {
  case 0xA3: // data port
    if ( DISK.cmd == DIVIDE_READ_BUFF ) {
      if ( DISK.buffer_ptr < DISK.buffer_len ) {
        data = DISK.buffer[DISK.buffer_ptr];
        DISK.buffer_ptr++;

        // readed whole buffer?
        if (DISK.buffer_ptr >= DISK.buffer_len)
          divide_in_port_BF = 0x40;
        else
          divide_in_port_BF = 0x08;
        return data;
      } else {
        divide_in_port_A7 = 0x04;
        divide_in_port_BF = 0x01;
        DISK.buffer_ptr = 0;
        DISK.buffer_len = 0;
        DISK.cmd = 0;
      }
    }
    return 0;
    break;
  case 0xA7: // error register
    data = divide_in_port_A7;
    divide_in_port_A7 = 0;
    return data;
    break;
  case 0xAB: // sector count register
    data = divide_port_AB;
    divide_port_AB = 0;
    return data;
    break;
  case 0xBF: // status register
    data = divide_in_port_BF;
    divide_in_port_BF = 0;
    return data;
    break;
  case 0xBB: // drive/head register
    return divide_port_BB;
    break;
  case 0xB7: // cylinder register high
    return divide_port_B7;
    break;
  case 0xB3: // cylinder register low
    return divide_port_B3;
    break;
  case 0xAF: // sector number register
    return divide_port_AF;
    break;
  default: // unknown divide port?
//    printf("  unknown port #%02X\n", portl);
    ;
  }
  return -1;
}

// OUT wrapper
void divide_port_out(int portl, byte data) {
  int i,drive;
  int lba;
  long size;
  
  drive = divide_active_drive;
  switch (portl) {
  case 0xE3: // divide_control_register
    // this port contains CONMEM, MAPRAM, x, x, x, x, RAM_BANK (2 bytes)
    divide_conmem = (data >> 7) & 1;
    // MAPRAM cannot be cleared (1->0)
    if ((data & 64)) {
      divide_mapram = 1;
    }
    divide_bank_changed = (divide_bank != (data & 0x03));
    divide_bank = (data & 0x03);
    break;
  case 0xBB: // drive/head register
    divide_active_drive = ((data >> 4) & 1);
    divide_port_BB = data;
    break;
  case 0xB7: // cylinder register high
    divide_port_B7 = data;
    break;
  case 0xB3: // cylinder register low
    divide_port_B3 = data;
    break;
  case 0xAF: // sector number register
    divide_port_AF = data;
    break;
  case 0xAB: // sector count register
    divide_port_AB = data;
    break;
  case 0xBF: // command register
    switch (data) {
      // silently ignore these commands
      case 0x00: // nop
      case 0x40: // read verify sectors with rerty
      case 0x41: // read verify sectors without rerty
      case 0x94: // standby immediate
      case 0x95: // idle immediate
      case 0x96: // standby
      case 0x97: // idle
      case 0x98: // check power mode
      case 0x99: // set sleep mode
      case 0xE7: // flush cache
        break;

      case 0x91: // initialise drive parameters
        if ((DISK.mbr_fd == -1) || (DISK.partition_fd == -1)) {
          // error?
          divide_in_port_BF &= 56;
          divide_in_port_BF |= 21;
          break;
        }
        // recalculate new geometry to fit CHS configuration
        size = DISK.h * DISK.s * DISK.c;
        DISK.h = (divide_port_BB&15)+1;
        DISK.s = divide_port_AB;
        DISK.c = size / (DISK.h * DISK.s);
        DISK.identify[108] = DISK.c & 0xFF;
        DISK.identify[109] = DISK.c >> 8;
        DISK.identify[110] = DISK.h & 0xFF;
        DISK.identify[111] = DISK.h >> 8;
        DISK.identify[112] = DISK.s & 0xFF;
        DISK.identify[113] = DISK.s >> 8;
        
        divide_in_port_BF &= 0x56;
        break;

      case 0xA1: // identify drive (ATAPI)
      case 0xEC: // identify drive (ATA)
        if ((DISK.mbr_fd == -1) || (DISK.partition_fd == -1)) {
          divide_in_port_BF &= 56;
          divide_in_port_BF |= 21;
          break;
        }
        memcpy(DISK.buffer, DISK.identify, 512);
        DISK.buffer_ptr = 0;
        DISK.buffer_len = 512;
        divide_in_port_BF &= 0x16;
        divide_in_port_BF |= 0x48;
        DISK.cmd=DIVIDE_READ_BUFF;
        break;

      case 0x20: // read sectors with retry
      case 0x21: // read sectors without retry
      case 0x22: // read long with retry
      case 0x23: // read long without retry
      case 0xC4: // read multiples
      case 0xC8: // read DMA with retry
      case 0xC9: // read DMA without retry

        if ((DISK.mbr_fd == -1) || (DISK.partition_fd == -1) || (divide_port_AB>DIVIDE_MAX_SEC_NR)) {
          divide_in_port_BF&=56;
          divide_in_port_BF|=21;
          break;
        }
        
        // LBA or CHS?
        if (divide_port_BB & 0x40)
          lba = ((divide_port_BB & 0x0F) << 24) | (divide_port_B7 << 16) | (divide_port_B3 << 8) | divide_port_AF; 
        else
          lba=chs2lba(drive, (divide_port_B7 << 8) | divide_port_B3, divide_port_BB & 0x0F, divide_port_AF);

        DISK.buffer_ptr = 0;
        DISK.buffer_len = 0;
        
        // send port_AB sectors
        for (i=0; i<divide_port_AB; i++)
        {
          // read 1 sector only. we cannot read all sectors at once because we have disk splitted into blocks
          if (lba < DISK.mbr_len) {
            DISK.mbr_ptr = lseek(DISK.mbr_fd, (lba << 9), SEEK_SET);
            read(DISK.mbr_fd, DISK.buffer + DISK.buffer_len, 512);
          } else {
            DISK.partition_ptr = lseek(DISK.partition_fd, ((lba - DISK.mbr_len) << 9), SEEK_SET);
            read(DISK.partition_fd, DISK.buffer + DISK.buffer_len, 512);
          }
          DISK.buffer_len += 512;
          lba++;
        }
        divide_in_port_BF &= 0x16;
        divide_in_port_BF |= 0x48;
        DISK.cmd = DIVIDE_READ_BUFF;
        break;
      case 0x30: // write sectors with retry
      case 0x31: // write sectors without retry
      case 0x32: // write long with retry
      case 0x33: // write long without retry
        DISK.buffer_ptr = 0;
        DISK.buffer_len = 0;
        DISK.cmd = DIVIDE_WRITE_BUFF;
        divide_in_port_BF &= 0x16;
        divide_in_port_BF |= 0x48;
        break;
      default:
//        printf("  unknown command #%02X\n", data);
        ;
    }
    break;
  case 0xA3: // data register
    if (DISK.mbr_fd==-1 || DISK.partition_fd==-1 || divide_port_AB>DIVIDE_MAX_SEC_NR) {
      break;
    }
    if (DISK.cmd!=DIVIDE_WRITE_BUFF)
      break;
    // store a byte into buffer
    memcpy(DISK.buffer + DISK.buffer_len, &data, 1);
    DISK.buffer_len++;
    
    // if is stored whole sector write it to disk
    if (DISK.buffer_len == 512) {

      // LBA or CHS?
      if (divide_port_BB & 0x40)
        lba = ((divide_port_BB & 0x0F) << 24) | (divide_port_B7 << 16) | (divide_port_B3 << 8) | divide_port_AF; 
      else
        lba = chs2lba(drive, (divide_port_B7 << 8) | divide_port_B3, divide_port_BB & 0x0F, divide_port_AF);
      
      // write sector to disk
      if (lba < DISK.mbr_len) {
        DISK.mbr_ptr = lseek(DISK.mbr_fd, (lba << 9), SEEK_SET);
        write(DISK.mbr_fd, DISK.buffer + DISK.buffer_ptr, 512);
      } else {
        DISK.partition_ptr = lseek(DISK.partition_fd, ((lba - DISK.mbr_len) << 9), SEEK_SET);
        write(DISK.partition_fd, DISK.buffer + DISK.buffer_ptr, 512);
      }
      DISK.buffer_ptr += 512;
      divide_port_AB--;
      lba++;
    }
    // we are done with writing?
    if (!divide_port_AB) {
      DISK.cmd=0;
      divide_in_port_BF=0x04;
    }
    break;
  default:
//    printf("  unknown port #%02X\n", portl);
      ;
  }
}


void divide_mapper(int addr) {
  // "M1 state"
  // mapper logic, see divIDE documentation for more info
  
  // check for divide memory connecting entry points
  if ((addr == 0x0000) ||
      (addr == 0x0008) ||
      (addr == 0x0038) ||
      (addr == 0x0066) ||
      (addr == 0x04C6) ||
      (addr == 0x0562)) divide_automap = 1;
  
  // check for original memory connecting entry points
  if ((addr & 0x1ff8) == 0x1ff8) {
    divide_automap = 0;
  }
  
  // if CONMEM is set, connect EEPROM (read only if JP2 is closed) and selected RAM bank (read/write)
  if (divide_conmem) { // map eeprom+ram
    if ((divide_memstate != 1) || divide_bank_changed) {
      memcpy(PRNM(proc).mem, divide_eeprom, 0x2000);
      memcpy(PRNM(proc).mem + 0x2000, divide_ram[divide_bank], 0x2000);
    }
    divide_memstate = 1;
    divide_bank_changed = 0;
    return;
  }
  
  // if CONMEM isn't set and MAPRAM is set and:
  //   1) if we reached "connecting" entry point, connect RAM #3 bank (read only) and selected RAM bank (if it is 3 it is too read only, else it is writable)
  //   2) if we reached "disconnecting" entry point, connect original ZX Speccy ROM
  if (divide_mapram) {
    if (divide_automap) { // map ram3+ram
      if ((divide_memstate != 2) || divide_bank_changed) {
        memcpy(PRNM(proc).mem, divide_ram[3], 0x2000);
        memcpy(PRNM(proc).mem + 0x2000, divide_ram[divide_bank], 0x2000);
      }
      divide_memstate = 2;
      divide_bank_changed = 0;
    } else { // map original_rom
      if (divide_memstate) {
        memcpy(PRNM(proc).mem, orig_rom, 0x4000);
      }
      divide_memstate = 0;
      divide_bank_changed = 0;
    }
    return;
  }
  
  // if JP2 is closed and:
  //   1) if we reached "connecting" entry point, connect EEPROM (read only) and selected RAM bank (read/write)
  //   2) if we reached "disconnecting" entry point, connect original ZX Speccy ROM
  if (divide_jp2) {
    if (divide_automap) { // map eeprom+ram
      if ((divide_memstate != 1) || divide_bank_changed) {
        memcpy(PRNM(proc).mem, divide_eeprom, 0x2000);
        memcpy(PRNM(proc).mem + 0x2000, divide_ram[divide_bank], 0x2000);
      }
      divide_memstate = 1;
      divide_bank_changed = 0;
    } else { // map original_rom
      if (divide_memstate) {
        memcpy(PRNM(proc).mem, orig_rom, 0x4000);
      }
      divide_memstate = 0;
      divide_bank_changed = 0;
    }
    return;
  }
  
  // else connect original ZX Speccy ROM
  if (divide_memstate) { // map original_rom
    memcpy(PRNM(proc).mem, orig_rom, 0x4000);
  }
  divide_memstate = 0;
  divide_bank_changed = 0;
}

void divide_put_mem(int addr, byte *ptr, byte val) {
  switch (divide_memstate) {
    case 0:
      // original ZX Speccy ROM - no update memory
      break;
    case 1:
      // EEPROM + selected RAM bank
      if (addr<0x2000) {
        // is EEPROM writable?
        if (divide_jp2) break;
        *ptr=val;
        // we have to make this change in divide memory too (not really mapped, only copied)
        divide_eeprom[addr]=val;
        break;
      }
      *ptr=val;
      // we have to make this change in divide memory too (not really mapped, only copied)
      divide_ram[divide_bank][addr-0x2000]=val;
      break;
    case 2:
      // RAM #3 bank + selected RAM bank
      // if it is 0k-8k page, write is disabled
      // if it is 8k-16k page, write is disabled if is selected RAM #3 bank
      if (addr<0x2000 || divide_bank==3) break;
      *ptr=val;
      // we have to make this change in divide memory too (not really mapped, only copied)
      divide_ram[divide_bank][addr-0x2000]=val;
      break;
    
    default:
//      printf("unknown divide state %d\n", divide_memstate);
      ;
  }
}

// JP2 (EEPROM "defender") switch
void divide_switch_jp2() {
  
  divide_jp2 ^= 1;
  sprintf(msgbuf, "JP2 %s", divide_jp2 ? "fused" : "opened");
  put_msg(msgbuf);
}

